﻿Imports System.Runtime.CompilerServices

Public Module Extensions


   Public Structure ApproachResult
      ''' <summary> der gesuchte Pfad, in Segmente zerteilt </summary>
      Public ReadOnly PathSegments As IList(Of String)
      ''' <summary> der Treenode, der dem Pfad am besten entspricht - im Idealfall entspricht er vollständig, schlimmstenfalls ist er Nothing - wenn nichtmal das Root-Segment matcht </summary>
      Public ReadOnly BestMatch As TreeNode
      ''' <summary> die Nodes, in die der Rest des Pfades ggfs. einzuhängen ist. Selbst wenn der Pfad ühaupt nicht matcht, sind diese Nodes gültig - dann sinds die Treeview.Nodes</summary>
      Public ReadOnly Nodes As TreeNodeCollection
      ''' <summary> die Anzahl schritte, die auf dem Pfad gangbar waren </summary>
      Public ReadOnly StepsDone As Integer
      ''' <summary> der Einfüge-Index, wo in Nodes der Rest-Pfad einzuhängen wäre </summary>
      Public ReadOnly InsertIndex As Integer

      Public Sub New(Nodes As TreeNodeCollection, Node As TreeNode, Segments As IList(Of String), StepsDone As Integer, InsertIndex As Integer)
         Me.Nodes = Nodes
         Me.BestMatch = Node
         Me.PathSegments = Segments
         Me.StepsDone = StepsDone
         Me.InsertIndex = InsertIndex
      End Sub

      ''' <summary> hängt den Rest des Pfades, der nicht mehr gematcht hat, an die richtige Position ein, und gibt den letzten eingehängten Node zurück. InvalidOperationException, wenn der Pfad restlos gematcht hat </summary>
      Public Function InsertRest() As TreeNode
         If IsFullMatch Then Throw New InvalidOperationException("Es gibt keinen Rest zu inserten!")
         Dim nd = Me.Nodes.Insert(InsertIndex, PathSegments(StepsDone))
         For i = StepsDone + 1 To PathSegments.Count - 1
            nd = nd.Nodes.Add(PathSegments(i))
         Next
         Return nd
      End Function

      Public ReadOnly Property IsFullMatch() As Boolean
         Get
            Return InsertIndex < 0
         End Get
      End Property
   End Structure 'ApproachResult

   ''' <summary>
   ''' sucht in (sortierten) Nodes und SubNodes anhand eines Pfades nach dem bezeichneten Node.
   ''' Bei Nichtexistenz des ZielNode den übergeordneten, in den der ZielNode zu erstellen wäre
   ''' </summary>
   ''' <param name="PathSegments">Die Segmente des Pfades</param>
   ''' <returns>ein <seealso cref="ApproachResult"> ApproachResult</seealso>, 
   ''' das den gefundenen Node enthält, und die Anzahl abgearbeiteter Segmente
   ''' </returns>
   <Extension()> _
   Public Function Approach(Nodes As TreeNodeCollection, PathSegments As IList(Of String)) As ApproachResult
      Dim Nd As TreeNode = Nothing, I As Integer
      For I = 0 To PathSegments.Count - 1
         Dim indx = Nodes.FindIndexByText(PathSegments(I))
         If indx < 0 Then Return New ApproachResult(Nodes, Nd, PathSegments, I, Not indx)
         Nd = Nodes(indx)
         Nodes = Nd.Nodes
      Next
      Return New ApproachResult(Nodes, Nd, PathSegments, I, -1)
   End Function

   ''' <summary>
   ''' sucht in (sortierten) Nodes nach einer Übereinstimmung mit txt. Wenn gefunden, dessen Index returnen. Wird ein Node angetroffen, dessen .Text grösser ist als txt, so ist die Suche auch beendet, denn aufgrund der Sortierung kann im folgenden kein Treffer mehr auftreten. Der Index dieses Auftretens bezeichnet dann die **EinfügePosition**, also wo ein matchender Node einzufügen wäre. Dieser Index wird bitweise negiert returnt, damit er von einem richtigen Treffer am Vorzeichen unterscheidbar ist, aber dennoch bei Bedarf ausgewertet werden kann - durch nochmaliges invertieren.
   ''' Die Konvention, die Einfügeposition als invertierten Index zu returnen, entspricht der Konvention von BinarySearch. Tatsächlich wäre hier auch eine BinarySearch angemessen, aber meine Faulheit verhinderte das
   ''' </summary>
   <Extension()> _
   Public Function FindIndexByText(nodes As TreeNodeCollection, txt As String) As Integer
      For i = 0 To nodes.Count - 1
         Select Case String.Compare(nodes(i).Text, txt, True)
            Case Is > 0 : Return Not i
            Case 0 : Return i
            Case Else
         End Select
      Next
      Return Not nodes.Count
   End Function

   <Extension()> _
   Public Function Approach(TV As TreeView, NodePath As String) As ApproachResult
      Return TV.Nodes.Approach(NodePath.Split(TV.PathSeparator(0)))
   End Function

End Module
